/// <reference types="react" />
import { onEventType } from "./helpers/onEvent";
/** Basic abstract class for every component in web-ui-pack */
export default abstract class WUPBaseElement<Events extends WUP.Base.EventMap = WUP.Base.EventMap> extends HTMLElement {
    #private;
    /** Reference to global style element used by web-ui-pack */
    static get $refStyle(): HTMLStyleElement | null;
    static set $refStyle(v: HTMLStyleElement | null);
    /** StyleContent related to component */
    static get $style(): string;
    /** StyleContent related to component & inherrited components */
    static get $styleRoot(): string;
    /** Get unique id for html elements; Every getter returns new id */
    static get $uniqueId(): string;
    static get classNameHidden(): string;
    /** Options that applied to element */
    abstract $options: Record<string, any>;
    protected _opts: Record<string, any>;
    constructor();
    /** Returns true if element is appended (result of setTimeout on connectedCallback) */
    get $isReady(): boolean;
    /** Try to focus self or first possible children; returns true if succesful */
    focus(): boolean;
    /** Remove focus from element on any nested active element */
    blur(): void;
    /** Called when element is added to document */
    protected gotReady(): void;
    /** Called when element is removed from document */
    protected gotRemoved(): void;
    /** Called on Init and every time as options/attributes changed */
    protected gotChanges(propsChanged: Array<string> | null): void;
    /** Called when element isReady and at least one of observedOptions is changed */
    protected gotOptionsChanged(e: WUP.Base.OptionEvent): void;
    /** Called once on Init */
    protected gotRender(): void;
    /** Browser calls this method when the element is added to the document */
    protected connectedCallback(): void;
    /** Browser calls this method when the element is removed from the document;
     * (can be called many times if an element is repeatedly added/removed) */
    protected disconnectedCallback(): void;
    _isStopChanges: boolean;
    /** Called when element isReady and one of observedAttributes is changed */
    protected gotAttributeChanged(name: string, oldValue: string, newValue: string): void;
    /** Browser calls this method when attrs pointed in observedAttributes is changed */
    protected attributeChangedCallback(name: string, oldValue: string, newValue: string): void;
    dispatchEvent<K extends keyof Events>(type: K, eventInit?: EventInit): boolean;
    dispatchEvent(event: Event): boolean;
    addEventListener<K extends keyof Events>(type: K, listener: (this: WUPBaseElement, ev: Events[K]) => any, options?: boolean | AddEventListenerOptions): void;
    addEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | AddEventListenerOptions): void;
    removeEventListener<K extends keyof Events>(type: K, listener: (this: WUPBaseElement, ev: Events[K]) => any, options?: boolean | EventListenerOptions): void;
    removeEventListener(type: string, listener: EventListenerOrEventListenerObject, options?: boolean | EventListenerOptions): void;
    /** Calls dispatchEvent and returns created event */
    fireEvent<K extends keyof Events>(type: K, eventInit?: EventInit): Event;
    /** Array of removeEventListener() that called on remove */
    protected disposeLst: Array<() => void>;
    /** Add event listener and remove after component removed; @options.passive=true by default */
    appendEvent<T extends keyof E, K extends HTMLElement | Document, E extends HTMLElementEventMap & Record<string, Event>>(...args: Parameters<onEventType<T, K, E>>): () => void;
    /** Remove events/functions that was appended */
    protected dispose(): void;
    /** Returns true if el is instance of Node and contains pointer element */
    includes(el: unknown): boolean;
    /** Parse attribute and return result */
    getBoolAttr(attr: string, alt: boolean): boolean;
    getBoolAttr(attr: string, alt: boolean | undefined): boolean | undefined;
    /** Parse attribute and return result; if attr missed or invalid => returns this._opts[attr] */
    getNumAttr(attr: string): number | undefined;
    /** Returns value from window[key] according to [attr]="key" or $options[key] if attr is missed */
    getRefAttr<T>(attr: string): T | undefined;
    /**
     * Remove attr if value falseOrEmpty; set '' or 'true' if true for HTMLELement
     * @param isSetEmpty set if need '' instead of 'value'
     */
    setAttr(attr: string, v: boolean | string | undefined | null, isSetEmpty?: boolean): void;
}
declare global {
    namespace WUP.Base {
        /** Cast not-supported props to optional string */
        type toJSX<T> = {
            [P in keyof T]?: T[P] extends number | boolean | string | undefined ? T[P] : string;
        };
        type OptionEvent<T extends Record<string, any> = Record<string, any>> = {
            props: Array<Extract<keyof T, string>>;
            target: T;
        };
        type EventMap<T = HTMLElementEventMap> = HTMLElementEventMap & Record<keyof T, Event>;
        type JSXProps<T, Opts extends Record<string, any> = any> = React.DetailedHTMLProps<Omit<React.HTMLAttributes<T>, "className"> & {
            class?: string | undefined;
        }, T> & toJSX<Opts>;
    }
}
