import WUPBaseElement from "./baseElement";
import { animate } from "./helpers/animate";
import { mathScaleValue } from "./helpers/math";
const tagName = "wup-circle";
/** Arc/circle chart based on SVG
 * @example
 * <wup-circle
 *   back="true"
 *   from="0"
 *   to="360"
 *   space="2"
 *   min="0"
 *   max="100"
 *   width="14"
 *   corner="0.25"
 *   items="window.circleItems"
 *  ></wup-circle>
 * // or JS/TS
 * const el = document.createElement("wup-circle");
 * el.$options.items = [{value:20}]; // etc.
 * document.body.appendChild(el); */
export default class WUPCircleElement extends WUPBaseElement {
    #ctr = this.constructor;
    static get observedOptions() {
        return this.observedAttributes;
    }
    static get observedAttributes() {
        return ["items", "width", "back", "corner", "from", "to", "min", "max", "space"];
    }
    static get $styleRoot() {
        return `:root {
          --circle-0: var(--base-sep);
          --circle-1: var(--base-btn-bg);
          --circle-2: #ff9f00;
          --circle-3: #1fb13f;
          --circle-4: #9482bd;
          --circle-5: #8bc4d7;
          --circle-6: #1abdb5;
        }`;
    }
    static get $style() {
        return `${super.$style}
      :host {
        contain: style layout paint;
        display: block;
        position: relative;
        overflow: hidden;
        margin: auto;
        padding: 2px;
        --anim-time: 400ms;
      }
      :host>strong {
        display: block;
        position: absolute;
        transform: translate(-50%,-50%);
        top: 50%; left: 50%;
        font-size: larger;
      }
      :host>svg {
        overflow: visible;
        display: block;
      }
      :host>svg path {
        stroke-width: 0;
        fill-rule: evenodd;
      }
      :host>svg>path { fill: var(--circle-0); }
      :host>svg>g>path:nth-child(1) { fill: var(--circle-1); }
      :host>svg>g>path:nth-child(2) { fill: var(--circle-2); }
      :host>svg>g>path:nth-child(3) { fill: var(--circle-3); }
      :host>svg>g>path:nth-child(4) { fill: var(--circle-4); }
      :host>svg>g>path:nth-child(5) { fill: var(--circle-5); }
      :host>svg>g>path:nth-child(6) { fill: var(--circle-6); }`;
    }
    static $defaults = {
        width: 14,
        corner: 0.25,
        back: true,
        from: 0,
        to: 360,
        min: 0,
        max: 100,
        space: 2,
    };
    $options = {
        ...this.#ctr.$defaults,
        items: [],
    };
    $refSVG = this.make("svg");
    $refItems = this.make("g");
    $refLabel;
    _opts = this.$options;
    /** Creates new svg-part */
    make(tag) {
        return document.createElementNS("http://www.w3.org/2000/svg", tag);
    }
    gotChanges(propsChanged) {
        super.gotChanges(propsChanged);
        this._opts.items = this.getRefAttr("items") || [];
        this._opts.back = this.getBoolAttr("back", this._opts.back);
        ["width", "corner", "from", "to", "min", "max", "space"].forEach((key) => {
            this._opts[key] = this.getNumAttr(key);
        });
        if (propsChanged) {
            this.$refItems.textContent = "";
            this.$refSVG.textContent = ""; // clean before new render
        }
        this.gotRenderItems();
    }
    _animation;
    gotRenderItems() {
        this._animation?.stop(false);
        const angleMin = this._opts.from;
        let angleMax = this._opts.to;
        // render background circle
        if (this._opts.back) {
            const back = this.make("path");
            back.setAttribute("d", this.drawArc(angleMin, angleMax));
            this.$refSVG.appendChild(back);
        }
        // render items
        this.$refSVG.appendChild(this.$refItems);
        const { items, space } = this._opts;
        let vMin = this._opts.min ?? 0;
        let vMax = this._opts.max ?? 360;
        const animTime = Number.parseInt(getComputedStyle(this).getPropertyValue("--anim-time"), 10); // WARN: expected anim-time: 200ms
        if (items.length > 1) {
            vMin = 0;
            vMax = items.reduce((v, item) => item.value + v, 0);
            angleMax -= (items.length - (angleMax - angleMin === 360 ? 0 : 1)) * space;
        }
        let angleTo = 0;
        let angleFrom = angleMin;
        (async () => {
            for (let i = 0; i < items.length; ++i) {
                const a = items[i];
                const v = a.value;
                angleTo = mathScaleValue(v, vMin, vMax, angleMin, angleMax) + (angleFrom - angleMin);
                const path = this.$refItems.appendChild(this.make("path"));
                a.color && path.setAttribute("fill", a.color);
                const ms = items.length === 1 ? animTime : mathScaleValue(v, vMin, vMax, 0, animTime);
                const from = angleFrom;
                this._animation = animate(from, angleTo, ms, (animV) => {
                    path.setAttribute("d", this.drawArc(from, animV));
                });
                await this._animation.catch().finally(() => delete this._animation);
                angleFrom = angleTo + space;
            }
        })();
        // render/remove label
        let ariaLbl = "";
        if (items.length === 1) {
            this.$refLabel = this.$refLabel ?? this.appendChild(document.createElement("strong"));
            const rawV = items[0].value;
            const perc = Math.round(((angleTo - this._opts.from) * 100) / (this._opts.to - this._opts.from));
            this.renderLabel(this.$refLabel, perc, rawV);
            ariaLbl = this.$refLabel.textContent;
        }
        else {
            this.$refLabel && this.$refLabel.remove();
            delete this.$refLabel;
            ariaLbl = `Values: ${items.map((a) => a.value).join(",")}`;
        }
        this.$refSVG.setAttribute("aria-label", ariaLbl);
    }
    /** Called on every changeEvent */
    gotRender() {
        super.gotRender();
        this.$refSVG.setAttribute("viewBox", `0 0 100 100`);
        this.$refSVG.setAttribute("role", "img");
        this.appendChild(this.$refSVG);
    }
    /** Called every time as need text-value */
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    renderLabel(label, percent, rawValue) {
        label.textContent = `${percent}%`;
    }
    /** Returns svg-path for Arc according to options */
    drawArc(angleFrom, angleTo) {
        const r = 50;
        const center = [r, r];
        return drawArc(center, r, this._opts.width, angleFrom, angleTo, this._opts.corner);
    }
}
customElements.define(tagName, WUPCircleElement);
/** Returns svg-path for Cirle according to options */
export function drawCircle(center, r, width) {
    const inR = r - width;
    const [x, y] = center;
    return [
        `M${x - r} ${y}`,
        `A${r} ${r} 0 1 0 ${x + r} ${y}`,
        `A${r} ${r} 0 1 0 ${x - r} ${y}`,
        `M${x - inR} ${y}`,
        `A${inR} ${inR} 0 1 0 ${x + inR} ${y}`,
        `A${inR} ${inR} 0 1 0 ${x - inR} ${y}`,
        "Z",
    ].join(" ");
}
/** Returns x,y for point in the circle */
function pointOnArc(center, r, angle) {
    const radians = ((angle - 90) * Math.PI) / 180.0;
    return [center[0] + r * Math.cos(radians), center[1] + r * Math.sin(radians)];
}
/** Returns svg-path for Arc according to options */
export function drawArc(center, r, width, angleFrom, angleTo, cornerR) {
    if (Math.abs(angleTo - angleFrom) === 360) {
        return drawCircle(center, r, width);
    }
    const inR = r - width;
    const circumference = Math.abs(angleTo - angleFrom);
    cornerR = Math.min(width / 2, width * cornerR);
    if (360 * (cornerR / (Math.PI * (r - width))) > Math.abs(angleFrom - angleTo)) {
        cornerR = (circumference / 360) * inR * Math.PI;
    }
    // inner and outer radiuses
    const inR2 = inR + cornerR;
    const outR = r - cornerR;
    // butts corner points
    const oStart = pointOnArc(center, outR, angleFrom);
    const oEnd = pointOnArc(center, outR, angleTo);
    const iStart = pointOnArc(center, inR2, angleFrom);
    const iEnd = pointOnArc(center, inR2, angleTo);
    const iSection = 360 * (cornerR / (2 * Math.PI * inR));
    const oSection = 360 * (cornerR / (2 * Math.PI * r));
    // arcs endpoints
    const iArcStart = pointOnArc(center, inR, angleFrom + iSection);
    const iArcEnd = pointOnArc(center, inR, angleTo - iSection);
    const oArcStart = pointOnArc(center, r, angleFrom + oSection);
    const oArcEnd = pointOnArc(center, r, angleTo - oSection);
    const arcSweep1 = circumference > 180 + 2 * oSection ? 1 : 0;
    const arcSweep2 = circumference > 180 + 2 * iSection ? 1 : 0;
    return [
        `M ${oStart[0]} ${oStart[1]}`,
        `A ${cornerR} ${cornerR} 0 0 1 ${oArcStart[0]} ${oArcStart[1]}`,
        `A ${r} ${r} 0 ${arcSweep1} 1 ${oArcEnd[0]} ${oArcEnd[1]}`,
        `A ${cornerR} ${cornerR} 0 0 1 ${oEnd[0]} ${oEnd[1]}`,
        `L ${iEnd[0]} ${iEnd[1]}`,
        `A ${cornerR} ${cornerR} 0 0 1 ${iArcEnd[0]} ${iArcEnd[1]}`,
        `A ${inR} ${inR} 0 ${arcSweep2} 0 ${iArcStart[0]} ${iArcStart[1]}`,
        `A ${cornerR} ${cornerR} 0 0 1 ${iStart[0]} ${iStart[1]}`,
        "Z", // end path
    ].join(" ");
}
// example: https://medium.com/@pppped/how-to-code-a-responsive-circular-percentage-chart-with-svg-and-css-3632f8cd7705
// similar https://github.com/w8r/svg-arc-corners
// demo https://milevski.co/svg-arc-corners/demo/
