import nestedProperty from "../helpers/nestedProperty";
import onEvent from "../helpers/onEvent";
import promiseWait from "../helpers/promiseWait";
import WUPSpinElement from "../spinElement";
import { WUPcssIcon, WUPcssScrollSmall } from "../styles";
import WUPBaseComboControl from "./baseCombo";
/* c8 ignore next */
/* istanbul ignore next */
!WUPSpinElement && console.error("!"); // It's required otherwise import is ignored by webpack
const tagName = "wup-select";
/** Form-control with dropdown/combobox behavior
 * @example
  const el = document.createElement("wup-select");
  el.$options.name = "gender";
  el.$options.items = [
    { value: 1, text: "Male" },
    { value: 2, text: "Female" },
    { value: 3, text: "Other/Skip" },
  ];
  el.$initValue = 3;
  el.$options.validations = { required: true };
  const form = document.body.appendChild(document.createElement("wup-form"));
  form.appendChild(el);
  // or HTML
  <wup-form>
    <wup-select name="gender" initvalue="3" validations="myValidations" items="myDropdownItems" />
  </wup-form>;
 * @tutorial innerHTML @example
 * <label>
 *   <span> // extra span requires to use with icons via label:before, label:after without adjustments
 *      <input/>
 *      <strong>{$options.label}</strong>
 *   </span>
 *   <button clear/>
 *   <wup-popup menu>
 *      <ul>
 *          <li>Item 1</li>
 *          <li>Item 2</li>
 *          // etc/
 *      </ul>
 *   </wup-popup>
 * </label>
 */
export default class WUPSelectControl extends WUPBaseComboControl {
    /** Returns this.constructor // watch-fix: https://github.com/Microsoft/TypeScript/issues/3841#issuecomment-337560146 */
    #ctr = this.constructor;
    static get observedOptions() {
        const arr = super.observedOptions;
        arr.push("items");
        return arr;
    }
    static get $styleRoot() {
        return `:root {
        --ctrl-select-icon-img: url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' width='768' height='768'%3E%3Cpath d='m16.078 258.214 329.139 329.139c21.449 21.449 56.174 21.449 77.567 0l329.139-329.139c21.449-21.449 21.449-56.174 0-77.567s-56.174-21.449-77.567 0L384 471.003 93.644 180.647c-21.449-21.449-56.173-21.449-77.567 0s-21.449 56.173 0 77.567z'/%3E%3C/svg%3E");
        --ctrl-select-current: inherit;
        --ctrl-select-current-bg: #d9f7fd;
      }`;
    }
    // WARN: scroll sets for ul otherwise animation brokes scroll to selectItem because animation affects on scrollSize
    static get $style() {
        return `${super.$style}
      :host {
        --ctrl-icon-img: var(--ctrl-select-icon-img);
      }
      :host[opened] label::after {
        transform: rotate(180deg);
      }
      :host [menu] {
        overflow: hidden;
      }
      :host [menu] ul {
        margin: 0;
        padding: 0;
        list-style-type: none;
        cursor: pointer;
        overflow: auto;
        max-height: 300px;
      }
      ${WUPcssScrollSmall(":host [menu]>ul")}
      :host [menu] li {
        padding: 1em;
      }
      @media (hover: hover) {
        :host [menu] li:hover {
          color: var(--ctrl-select-current);
          background: var(--ctrl-select-current-bg);
        }
      }
      :host [menu] li[aria-selected="true"] {
        color: var(--ctrl-select-current);
        background: var(--ctrl-select-current-bg);
        display: flex;
      }
      :host [menu] li[aria-selected="true"]:after {
        content: "";
        --ctrl-icon-img: var(--wup-icon-check);
        ${WUPcssIcon}
        margin-left: auto;
      }
      :host [menu] li[focused] {
        box-shadow: inset 0 0 4px 0 var(--ctrl-focus);
        border-radius: var(--ctrl-border-radius);
      }`;
    }
    /** Text for listbox when no items are displayed */
    static $textNoItems = "No Items";
    /** Text for aria-label of <ul> element */
    static $ariaLabelItems = "Items";
    /** Function to filter menuItems based on inputValue; all values are in lowerCase */
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    static $filterMenuItem(menuItemText, inputValue, inputRawValue) {
        return menuItemText.startsWith(inputValue) || menuItemText.includes(` ${inputValue}`);
    }
    static $defaults = {
        ...WUPBaseComboControl.$defaults,
        validationRules: { ...WUPBaseComboControl.$defaults.validationRules },
        showCase: 4 /* ShowCases.onClick */ | 2 /* ShowCases.onFocus */ | 64 /* ShowCases.onPressArrowKey */ | 32 /* ShowCases.onInput */,
    };
    $options = {
        ...this.#ctr.$defaults,
        items: [],
    };
    _opts = this.$options;
    /** Returns whether control in pending state or not (shows spinner) */
    get $isPending() {
        return !!this.#stopPending;
    }
    /** Called when need to parse attr [initValue] */
    parse(attrValue) {
        this.getItems().then((arr) => {
            let r;
            if (arr?.length) {
                r =
                    attrValue === ""
                        ? arr.find((o) => o.value == null)
                        : arr.find((o) => o.value && `${o.value}` === attrValue);
            }
            this._noDelInitValueAttr = true;
            this.$initValue = r?.value;
            delete this._noDelInitValueAttr;
        });
        return this.$initValue;
    }
    /** Called on every spin-render */
    renderSpin() {
        const spin = document.createElement("wup-spin");
        spin.$options.fit = true;
        spin.$options.overflowFade = true;
        spin.$options.overflowTarget = this;
        this.appendChild(spin);
        return spin;
    }
    async gotOptionsChanged(e) {
        const ev = e;
        if (ev.props.includes("items")) {
            this.removePopup();
            this._cachedItems = undefined;
            /* istanbul ignore else */
            if (ev.props.length === 1) {
                return; // skip re-init if only $options.items is changed
            }
        }
        super.gotOptionsChanged(e);
    }
    gotFormChanges(propsChanged) {
        super.gotFormChanges(propsChanged);
        this.$refInput.readOnly = this.$refInput.readOnly || this.$isPending;
    }
    #stopPending;
    /** Change pending state */
    changePending(v) {
        if (v === !!this.#stopPending) {
            return;
        }
        if (v) {
            this.$refInput.setAttribute("aria-busy", true);
            const wasDisabled = this._opts.disabled;
            const refSpin = this.renderSpin();
            this.$refInput.readOnly = true;
            this.#stopPending = () => {
                this.$refInput.readOnly = this.$isReadOnly;
                this.#stopPending = undefined;
                this.$options.disabled = wasDisabled;
                this.$refInput.removeAttribute("aria-busy");
                refSpin.remove();
            };
        }
        else {
            this.#stopPending();
        }
    }
    /** All items of current menu */
    _menuItems;
    /** Items resolved from options */
    _cachedItems;
    /** Called when NoItems need to show */
    renderMenuNoItems(popup, isReset) {
        if (isReset) {
            popup.hidden = false;
            const liSaved = this._menuItems._refNoItems;
            if (liSaved) {
                liSaved.style.display = "none";
            }
        }
        else if (this.#ctr.$textNoItems) {
            const liSaved = this._menuItems._refNoItems;
            if (liSaved) {
                liSaved.style.display = "";
            }
            else {
                const ul = popup.children.item(0);
                const li = ul.appendChild(document.createElement("li"));
                li.textContent = this.#ctr.$textNoItems;
                li.setAttribute("role", "option");
                li.setAttribute("aria-disabled", true);
                li.setAttribute("aria-selected", false);
                this._menuItems._refNoItems = li;
            }
        }
        else {
            popup.hidden = true;
        }
    }
    async renderMenu(popup, menuId) {
        const ul = popup.appendChild(document.createElement("ul"));
        ul.setAttribute("id", menuId);
        ul.setAttribute("role", "listbox");
        ul.setAttribute("aria-label", this.#ctr.$ariaLabelItems);
        const all = await this.renderMenuItems(ul);
        this._menuItems = { all, focused: -1 };
        !all.length && this.renderMenuNoItems(popup, false);
        // it happens on every show because by hide it dispose events
        onEvent(ul, "click", (e) => {
            e.stopPropagation(); // to prevent popup-hide-show
            let t = e.target;
            while (1) {
                const parent = t.parentElement;
                /* istanbul ignore else */
                if (parent === ul) {
                    this.gotMenuItemClick(e, t);
                    break;
                }
                t = parent;
            }
        });
        if (this._needFilter) {
            this._needFilter();
            delete this._needFilter;
        }
        return ul;
    }
    /** Method retrieves items from options and shows spinner if required */
    async getItems() {
        if (this._cachedItems) {
            return this._cachedItems;
        }
        const items = nestedProperty.get(window, this.getAttribute("items") || "") || this._opts.items;
        let arr;
        if (typeof items === "function") {
            const f = items();
            if (f instanceof Promise) {
                arr = await promiseWait(f, 300, (v) => this.changePending(v));
            }
            else {
                arr = f;
            }
        }
        else {
            arr = items;
        }
        this._cachedItems = arr;
        return arr;
    }
    // Called when need to show text related to value
    valueToText(v, items) {
        const i = items.findIndex((o) => this.#ctr.$isEqual(o.value, v));
        if (i === -1) {
            if (this._opts.allowNewValue) {
                return v != null ? v.toString() : "";
            }
            console.error(`${this.tagName}${this._opts.name ? `[${this._opts.name}]` : ""}. Not found in items`, {
                items,
                value: v,
            });
            return `Error: not found for ${v != null ? v.toString() : ""}`;
        }
        const item = items[i];
        if (typeof item.text === "function") {
            const li = document.createElement("li");
            const s = item.text(item.value, li, i);
            li.remove();
            return s;
        }
        return item.text;
    }
    valueToInput(v) {
        if (v === undefined) {
            return "";
        }
        const r = this.getItems().then((items) => this.valueToText(v, items));
        return r;
    }
    /** It's called with option allowNewValue to find value related to text */
    findValueByText(txt) {
        const s = txt.toLowerCase();
        const i = this._menuItems?.all.findIndex((o) => o._text === s);
        const r = i > -1 ? this._cachedItems[i].value : undefined;
        return r;
    }
    /** Create menuItems as array of HTMLLiElement with option _text required to filtering by input (otherwise content can be html-structure) */
    async renderMenuItems(ul) {
        const arr = await this.getItems();
        const arrLi = arr.map(() => {
            const li = ul.appendChild(document.createElement("li"));
            li.setAttribute("role", "option");
            // li.setAttribute("aria-selected", "false");
            return li;
        });
        if (arr.length) {
            if (typeof arr[0].text === "function") {
                arr.forEach((v, i) => {
                    arrLi[i]._text = v.text(v.value, arrLi[i], i).toLowerCase();
                });
            }
            else {
                arr.forEach((v, i) => {
                    const txt = v.text;
                    arrLi[i].textContent = txt;
                    arrLi[i]._text = txt.toLowerCase();
                });
            }
        }
        return arrLi;
    }
    /** Called when need to setValue & close base on clicked item */
    gotMenuItemClick(_e, item) {
        const i = this._menuItems.all.indexOf(item);
        const o = this._cachedItems[i];
        this.selectValue(o.value);
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    async goShowMenu(showCase, e, isNeedWait) {
        if (this.$isPending) {
            return null;
        }
        if (this.$refPopup && showCase !== 32 /* ShowCases.onInput */ && this._menuItems.filtered) {
            this._menuItems.all.forEach((li) => (li.style.display = "")); // reset styles after filtering
            delete this._menuItems.filtered;
            this.renderMenuNoItems(this.$refPopup, true); // remove NoItems
        }
        const popup = await super.goShowMenu(showCase, e, isNeedWait);
        if (!popup) {
            return null;
        }
        // set aria-selected
        this.selectMenuItemByValue(this.$value);
        return popup;
    }
    /** Select menu item if value !== undefined and menu is opened */
    selectMenuItemByValue(v) {
        if (v !== undefined && this.$isOpen) {
            const i = this._cachedItems.findIndex((item) => this.#ctr.$isEqual(item.value, v));
            this.selectMenuItem(this._menuItems.all[i] || null);
        }
    }
    focusMenuItem(next) {
        next && !next.hasAttribute("aria-selected") && next.setAttribute("aria-selected", false);
        super.focusMenuItem(next);
        if (next === null && this._menuItems) {
            this._menuItems.focused = -1;
        }
    }
    /** Focus item by index or reset is index is null (via aria-activedescendant).
     *  If menuItems is filtered by input-text than index must point on filtered array */
    focusMenuItemByIndex(index) {
        const { filtered } = this._menuItems;
        const trueIndex = filtered ? filtered[index] : index;
        const next = this._menuItems.all[trueIndex];
        this.focusMenuItem(next);
        this._menuItems.focused = index;
    }
    async gotKeyDown(e) {
        // pending event disables gotKeyDown so it's case impossible
        // if (this.$isPending) {
        //   return;
        // }
        const wasOpen = this.$isOpen;
        if (wasOpen && this._opts.allowNewValue && !this._focusedMenuItem && e.key === "Enter") {
            const txt = this.$refInput.value.trim();
            const v = txt ? this.findValueByText(txt) ?? txt : undefined;
            // this.selectValue(v);
            // e.preventDefault();
            // return;
            this.setValue(v);
        }
        await super.gotKeyDown(e);
        if (!this.$isOpen || e.altKey || e.shiftKey || e.ctrlKey || !this._menuItems) {
            return;
        }
        const { length } = this._menuItems.filtered || this._menuItems.all;
        let focusIndex = null;
        // firstFocused can be selected/current
        if (this._menuItems.focused === -1 && this._selectedMenuItem) {
            if (e.key === "ArrowDown" || e.key === "ArrowUp") {
                focusIndex = this._menuItems.all.indexOf(this._selectedMenuItem);
                // case impossible because with filtered items we have this._menuItems!.focused !== -1
                // if (this._menuItems.filtered) {
                //   focusIndex = this._menuItems.filtered.indexOf(focusIndex);
                // }
                // if (focusIndex === -1) {
                //   focusIndex = null;
                // }
            }
        }
        if (focusIndex === null) {
            switch (e.key) {
                case "ArrowDown":
                    {
                        let cur = this._menuItems.focused;
                        focusIndex = ++cur >= length ? 0 : cur;
                    }
                    break;
                case "ArrowUp":
                    {
                        let cur = wasOpen ? this._menuItems.focused : length;
                        focusIndex = --cur < 0 ? length - 1 : cur;
                    }
                    break;
                case "Home":
                    focusIndex = 0;
                    break;
                case "End":
                    focusIndex = length - 1;
                    break;
                default:
                    break;
            }
        }
        if (focusIndex != null) {
            e.preventDefault();
            this.focusMenuItemByIndex(focusIndex);
        }
    }
    /** For case when need to menu is opened but items are not rendered yet */
    _needFilter;
    gotInput(e) {
        this.$isOpen && this.focusMenuItem(null); // reset virtual focus
        super.gotInput(e);
        const rawV = e.target.value;
        const v = rawV.trimStart().toLowerCase();
        delete this._needFilter;
        const filter = () => {
            const filtered = [];
            this._menuItems.all.forEach((li, i) => {
                const isOk = !v || this.#ctr.$filterMenuItem(li._text, v, rawV);
                isOk && filtered.push(i);
                li.style.display = isOk ? "" : "none";
            });
            this._menuItems.filtered = filtered;
            const hasVisible = filtered.length !== 0;
            this.renderMenuNoItems(this.$refPopup, hasVisible);
            hasVisible && rawV !== "" && !this._opts.allowNewValue && this.focusMenuItemByIndex(0);
        };
        if (!this._menuItems) {
            this._needFilter = filter;
        }
        else {
            filter();
        }
        // this.$refPopup!.$refresh(); // it doesn't required anymore because popup listens for own sizes
    }
    gotFocusLost() {
        if (this._opts.allowNewValue) {
            // to allow user left value without press Enter
            const txt = this.$refInput.value.trim();
            const v = txt ? this.findValueByText(txt) ?? txt : undefined;
            this.setValue(v);
        }
        super.gotFocusLost();
    }
    setValue(v, canValidate = true, skipInput = false) {
        const r = super.setValue(v, canValidate, skipInput);
        r && this.selectMenuItemByValue(v);
        return r;
    }
    removePopup() {
        super.removePopup();
        this._menuItems = undefined;
    }
}
customElements.define(tagName, WUPSelectControl);
