import MaskTextInput from "./text.mask";
import WUPBaseControl from "./baseControl";
declare const tagName = "wup-text";
declare global {
    namespace WUP.Text {
        interface EventMap extends WUP.BaseControl.EventMap {
        }
        interface ValidityMap extends WUP.BaseControl.ValidityMap {
            /** If textLength < pointed shows message 'Min length is {x} characters` */
            min: number;
            /** If textLength > pointed shows message 'Max length is {x} characters` */
            max: number;
            /** If $value doesn't match email-pattern shows message 'Invalid email address` */
            email: boolean;
            /** If input value doesn't completely fit pointed mask shows pointed message
             * @default "Incomplete value"
             * @Rules
             * * enabled by default with $options.mask
             * * excluded from listing (for $options.validationShowAll)
             * * ignores control value, instead it uses `this.refMask` state based on `$refInput.value`
             *  */
            _mask: string;
            /** If parseInput() throws exception during the input-change is wrong then pointed message shows
             * @default "Invalid value"
             * @Rules
             * * processed only by input change (not value-change)
             * * removed by focusout (because input rollback to previous valid value)
             * * excluded from listing (for $options.validationShowAll) */
            _parse: string;
        }
        interface Defaults<T = string, VM = ValidityMap> extends WUP.BaseControl.Defaults<T, VM> {
            /** Debounce time to wait for user finishes typing to start validate and provide $change event
             * @defaultValue 0; */
            debounceMs?: number;
            /** Select whole text when input got focus (when input is not readonly and not disabled);
             * @defaultValue true */
            selectOnFocus: boolean;
            /** Show/hide clear button. @see ClearActions
             * @defaultValue true */
            clearButton: boolean;
        }
        interface Options<T = string, VM = ValidityMap> extends WUP.BaseControl.Options<T, VM>, Defaults<T, VM> {
            /** Make input masked
             * @rules when mask is pointed and contains only numeric vars
             * * inputmode='numeric' so mobile device show numeric-keyboard
             * * enables validation 'mask' with error message 'Incomplete value'
             * @example
             * "0000-00-00" // for date in format yyyy-mm-dd
             * "##0.##0.##0.##0" // IPaddress
             * "+1(000) 000-0000" // phoneNumber
             * "00:00 /[AP]/M" // time hh:mm AM/PM
             * '0' // required digit
             * '#' // optional digit
             * '*' // any char
             * '*{1,5}' // - any 1..5 chars
             * '//[a-zA-Z]//' // regex: 1 letter (WARN: regex must be pointed for checkin only 1 char at once)
             * '//[a-zA-Z]//{1,5}' // regex: 1..5 letters
             * '|0' // or '\x00' - static char '0'
             * '|#' // or '\x01' - static char '#'
             * '|*' // or '\x02' - static char '*'
             * '|/' // or '\x03' - static char '/'
             * */
            mask?: string;
            /** Placeholder for mask. By default it inherits from mask. To disabled it set 'false' or '' (empty string);
             *  for date maskholder can be 'yyyy-mm-dd' */
            maskholder?: string | false;
            /** Part before input; for example for value "$ 123 USD" prefix is "$ " */
            prefix?: string;
            /** Part after input; for example for value "$ 123 USD" prefix is " USD" */
            postfix?: string;
        }
        interface Attributes extends WUP.BaseControl.Attributes, Pick<Options, "mask" | "maskholder" | "prefix" | "postfix"> {
        }
        interface JSXProps<C = WUPTextControl> extends WUP.BaseControl.JSXProps<C>, Attributes {
        }
        interface GotInputEvent extends InputEvent {
            target: HTMLInputElement;
        }
    }
    interface HTMLElementTagNameMap {
        [tagName]: WUPTextControl;
    }
    namespace JSX {
        interface IntrinsicElements {
            [tagName]: WUP.Text.JSXProps;
        }
    }
}
/** Form-control with text-input
 * @example
  const el = document.createElement("wup-text");
  el.$options.name = "email";
  el.$options.validations = { required: true, max: 100, email: true };
  const form = document.body.appendChild(document.createElement("wup-form"));
  form.appendChild(el);
  // or HTML
  <wup-form>
    <wup-text name="firstName" initvalue="Donny" validations="myValidations"/>
  </wup-form>;
 * @tutorial innerHTML @example
 * <label>
 *   <span> // extra span requires to use with icons via label:before, label:after without adjustments
 *      <input/>
 *      <strong>{$options.label}</strong>
 *   </span>
 *   <button clear/>
 * </label>
 */
export default class WUPTextControl<ValueType = string, EventMap extends WUP.Text.EventMap = WUP.Text.EventMap> extends WUPBaseControl<ValueType, EventMap> {
    #private;
    static get observedOptions(): Array<string>;
    static get observedAttributes(): Array<string>;
    static get $styleRoot(): string;
    static get $style(): string;
    /** Default options - applied to every element. Change it to configure default behavior */
    static $defaults: WUP.Text.Defaults;
    $options: WUP.Text.Options<string>;
    protected _opts: WUP.Text.Options<string, WUP.Text.ValidityMap>;
    $refBtnClear?: HTMLButtonElement;
    $refMaskholder?: HTMLSpanElement;
    $refPrefix?: HTMLSpanElement;
    $refPostfix?: HTMLSpanElement;
    constructor();
    parse(text: string): ValueType | undefined;
    /** Called before parseInput on gotInput event */
    canParseInput(_text: string): boolean;
    /** Called to parse input text to value (related to locale or pointed format) */
    parseInput(text: string): ValueType | undefined;
    /** Returns true if need to use custom undo/redo (required when input somehow formatted/masked) */
    protected canHandleUndo(): boolean;
    protected get validations(): WUP.Text.Options["validations"];
    protected renderControl(): void;
    /** Create & append element to control */
    protected renderBtnClear(): HTMLButtonElement;
    /** Add/update/remove prefix part */
    protected renderPrefix(text: string | undefined): void;
    /** Add/update or remove prefix part */
    protected renderPostfix(text: string | undefined): void;
    protected gotFocus(): Array<() => void>;
    protected gotFocusLost(): void;
    protected gotChanges(propsChanged: Array<keyof WUP.Text.Options> | null): void;
    protected gotKeyDown(e: KeyboardEvent): void;
    /** Handler of 'beforeinput' event */
    protected gotBeforeInput(e: WUP.Text.GotInputEvent): void;
    /** Called when user types text OR when need to apply/reset mask (on focusGot, focusLost) */
    protected gotInput(e: WUP.Text.GotInputEvent): void;
    /** Mask object to proccess mask on input */
    refMask?: MaskTextInput;
    /** Called to apply mask-behavior (on "input" event) */
    protected maskInputProcess(e: WUP.Text.GotInputEvent | null): string;
    /** Add/update maskholder or skip if it's not defined */
    private renderMaskHolder;
    /** Convert values to history-snapshot; required for undo/redo logic of input */
    private historyToSnapshot;
    /** Parse history-snapshot; required for undo/redo logic of input */
    private historyFromSnapshot;
    _histUndo?: Array<string>;
    _histRedo?: Array<string>;
    /** Undo/redo input value (only if canHandleUndo() === true)
     * @returns true if action succeed (history not empty) */
    historyUndoRedo(toNext: boolean): boolean;
    /** Make undo for input after 100ms when user types not allowed chars
     * @tutorial Troubleshooting
     * * declineInput doesn't trigger beforeinput & input events (do it manually if required) */
    protected declineInput(nextCursorPos?: number): void;
    protected setValue(v: ValueType | undefined, canValidate?: boolean, skipInput?: boolean): boolean | null;
    /** Called to update/reset value for <input/> */
    protected setInputValue(v: ValueType | undefined | string): void;
    _onceErrName?: string;
    /** Called when inputValue != $value and need to show error on the fly by input-change */
    protected validateOnce(rule: {
        [key: string]: boolean | string | ((v: any) => false | string);
    }, force?: boolean): void;
    protected goHideError(): void;
}
export {};
