import { ShowCases, HideCases } from "./popupElement.types";
import WUPBaseElement from "../baseElement";
import WUPPopupArrowElement from "./popupArrowElement";
/** PopupElement
 * @example
 * JS/TS
 * ```js
 * WUPPopupElement.$defaults.arrowEnable = true;
 *
 * const el = document.createElement('wup-popup');
 * el.$options.showCase = ShowCases.onClick | ShowCases.onFocus;
 * el.$options.target = document.querySelector('button');
 * // if placement impossible according to rules rest of possible rules will be applied
 * el.$options.placement = [
 *  WUPPopupElement.$placements.$top.$middle,
 *  WUPPopupElement.$placements.$bottom.$middle,
 *  WUPPopupElement.$placements.$bottom.$middle.$adjust, // adjust means 'ignore align to fit layout`
 *  WUPPopupElement.$placements.$bottom.$middle.$adjust.$resizeHeight, // resize means 'allow to resize to fit layout'
 * ];
 * document.body.append(el);
 * // or
 * const btn = document.querySelector('button');
 * // this is the most recomended way because $attach appends popup only by onShow and removes byHide
 * const detach = WUPPopupElement.$attach(
                    { target: btn, text: "Some text content here", showCase: ShowCases.onFocus | ShowCases.onClick },
                    (popup) => { popup.className = "popup-class-here"; }
                  )'
 *```
 * HTML
 * ```html
 * <button id="btn1">Target</button>
 * <!-- You can skip pointing attribute 'target' if popup appended after target -->
 * <wup-popup target="#btn1" placement="top-start">Some content here</wup-popup>
 * ```
 * @tutorial Troubleshooting:
 * * You can set minWidth, minHeight to prevent squizing of popup or don't use rule '.$adjust'
 * * Don't override styles: display, transform (possible to override only for animation)
 * * Don't use inline styles: maxWidth, maxHeight
 * * If target removed (when popup $isOpen) and appended again you need to update $options.target (because $options.target cleared)
 * * Popup has overflow 'auto'; If you change to 'visible' it will apply maxWidth/maxHeight to first children (because popup must be restricted by maxSize to avoid layout issues)
 * * During the closing attr 'hide' is appended only if css-animation-duration is detected
 */
export default class WUPPopupElement<Events extends WUP.Popup.EventMap = WUP.Popup.EventMap> extends WUPBaseElement<Events> {
    #private;
    static get observedOptions(): Array<keyof WUP.Popup.Options>;
    static get observedAttributes(): Array<LowerKeys<WUP.Popup.Options>>;
    static $placements: {
        $top: WUP.Popup.Place.EdgeFunc;
        $bottom: WUP.Popup.Place.EdgeFunc;
        $left: WUP.Popup.Place.EdgeFunc;
        $right: WUP.Popup.Place.EdgeFunc;
    };
    static $placementAttrs: {
        "top-start": WUP.Popup.Place.AdjustFunc;
        "top-middle": WUP.Popup.Place.AdjustFunc;
        "top-end": WUP.Popup.Place.AdjustFunc;
        "bottom-start": WUP.Popup.Place.AdjustFunc;
        "bottom-middle": WUP.Popup.Place.AdjustFunc;
        "bottom-end": WUP.Popup.Place.AdjustFunc;
        "left-start": WUP.Popup.Place.AdjustFunc;
        "left-middle": WUP.Popup.Place.AdjustFunc;
        "left-end": WUP.Popup.Place.AdjustFunc;
        "right-start": WUP.Popup.Place.AdjustFunc;
        "right-middle": WUP.Popup.Place.AdjustFunc;
        "right-end": WUP.Popup.Place.AdjustFunc;
    };
    /** Default options. Change it to configure default behavior */
    static $defaults: Omit<WUP.Popup.Options, "target">;
    static get $styleRoot(): string;
    static get $style(): string;
    /** Listen for target according to showCase and create/remove popup when it's required (by show/hide).
     *  This helps to avoid tons of hidden popups on HTML;
     *  Firing detach doesn't required if target removed by target.remove() or target.parent.removeChild(target);
     *  If target is removed via changing innerHTML you should fire detach() to avoid memoryLeak
     *  @returns detach-function (hide,remove popup and remove eventListeners)
     *  @example
     *  const detach = WUPPopupElement.$attach(
     *     {
     *       target: document.querySelector("button") as HTMLElement,
     *       text: "Some text here",
     *       showCase: ShowCases.onClick,
     *     },
     *     // (el) => el.class = "popup-attached"
     *   );
     * @tutorial Troubleshooting:
     * * $attach doesn't work with showCase.always it doesn't make sense
     * * every new attach on the same target > re-init previous (1 attach per target is possible)
     * * Firing detach() doesn't required if target removed by `target.remove()` or `target.parent.removeChild(target)`;
     * * If popup is hidden and target is removed via `target.parent.innerHTML="another content"` you should fire detach() to avoid memoryLeak
     */
    static $attach<T extends WUPPopupElement>(options: WUP.Popup.AttachOptions, 
    /** Fires when popup is added to document */
    callback?: (el: T) => void): () => void;
    /** All options for this popup. If you want to change common options @see WUPPopupElement.$defaults */
    $options: WUP.Popup.Options;
    protected _opts: WUP.Popup.Options;
    /** Hide popup
     * @returns Promise resolved by animation time */
    $hide(): Promise<void>;
    /** Show popup; it disables $options.showCase and rollbacks by $hide().
     * @returns Promise resolved by animation time */
    $show(): Promise<void>;
    /** Force to update position when popup $isOpen. Call this if popup content is changed */
    $refresh(): void;
    /** Returns if popup is opened */
    get $isOpen(): boolean;
    /** Returns arrowElement if $options.arrowEnable=true and after popup $isOpen */
    get $refArrow(): WUPPopupArrowElement | null;
    protected gotReady(): void;
    /** Called after gotReady() and $show() (to reinit according to options) */
    protected init(): void;
    protected gotChanges(propsChanged: Array<string> | null): void;
    protected connectedCallback(): void;
    protected setMaxHeight(v: string): void;
    protected setMaxWidth(v: string): void;
    /** Collect/calc all required values into #state (when menu shows) */
    protected buildState(): void;
    /** Required to stop previous animations/timeouts (for case when option animation is changed) */
    _stopAnimation?: () => void;
    protected goAnimate(animTime: number, isClose: boolean): Promise<boolean>;
    /** Shows popup if target defined; returns true if successful */
    protected goShow(showCase: ShowCases): boolean | Promise<boolean>;
    _isClosing?: true;
    /** Hide popup. @hideCase as reason of hide(). Calling 2nd time at once will stop previous hide-animation */
    protected goHide(hideCase: HideCases): boolean | Promise<boolean>;
    protected gotRemoved(): void;
    /** Call when need to reinit */
    protected disposeListener(): void;
    protected dispose(): void;
}
declare const tagName = "wup-popup";
declare global {
    interface HTMLElementTagNameMap {
        [tagName]: WUPPopupElement;
    }
    namespace JSX {
        interface IntrinsicElements {
            [tagName]: WUP.Base.JSXProps<WUPPopupElement> & Partial<{
                /** QuerySelector to find target - anchor that popup uses for placement.
                 * If attr.target and $options.target are empty previousSibling will be attached.
                 * Popup defines target on show()
                 *
                 * attr `target` has hire priority than ref.options.target
                 *  */
                target: string;
                /** Placement rule (relative to target); applied on show(). Call show() again to apply changed options */
                placement: keyof typeof WUPPopupElement.$placementAttrs;
                /** @deprecated SyntheticEvent is not supported. Use ref.addEventListener('$show') instead */
                onShow: never;
                /** @deprecated SyntheticEvent is not supported. Use ref.addEventListener('$hide') instead */
                onHide: never;
                /** @deprecated SyntheticEvent is not supported. Use ref.addEventListener('$willHide') instead */
                onWillHide: never;
                /** @deprecated SyntheticEvent is not supported. Use ref.addEventListener('$willShow') instead */
                onWillShow: never;
                /** @readonly Result position; use this to restyle animation etc. */
                readonly position: "top" | "left" | "bottom" | "right";
                /** @readonly Hide state; use this to hide-animation */
                readonly hide: "";
            }>;
        }
    }
}
export {};
