import focusFirst from "../helpers/focusFirst";
import onEvent from "../helpers/onEvent";
import onFocusGot from "../helpers/onFocusGot";
import onFocusLost from "../helpers/onFocusLost";
/** Listen for target according to showCase. If target removed call stopListen + don't forget to remove popup yourself;
 * @tutorial Troubleshooting/rules:
 * * To allow user hide/show popup before previous action is ended provide for onShow and onHide-
 *   callbacks results as fast as it's possible without waiting for animation */
export default function popupListen(options, 
/** If succesfull callback must return HTMLElement */
onShow, onHide) {
    const opts = { ...popupListen.$defaults, ...options };
    const t = opts.target;
    if (!(t instanceof HTMLElement)) {
        throw new Error("WUP-Popup. Target is required");
    }
    if (!t.isConnected) {
        throw new Error("WUP-Popup. Target must be appended to layout");
    }
    let openedEl = null;
    /** Events that must be added on show; Should return removeCallbacks/onHideCallbacks */
    const onShowCallbacks = [];
    const onHideCallbacks = [];
    const onRemoveCallbacks = [];
    function onHideRef() {
        onHideCallbacks.forEach((f) => f());
        onHideCallbacks.length = 0;
    }
    function stopListen() {
        openedEl = null;
        onRemoveCallbacks.forEach((f) => f());
        onRemoveCallbacks.length = 0;
        onHideRef();
        onShowCallbacks.length = 0;
    }
    async function show(showCase, e) {
        if (openedEl || show._isDoing) {
            return;
        }
        show._isDoing = true;
        // eslint-disable-next-line no-async-promise-executor
        show._isDoing = new Promise(async (res, rej) => {
            try {
                openedEl = await onShow(showCase, e || null);
                res();
            }
            catch (err) {
                rej(err);
            }
        }).finally(() => (show._isDoing = false));
        await show._isDoing;
        // timeout required to avoid immediate hide by bubbling events to root
        openedEl && setTimeout(() => openedEl && onShowCallbacks.forEach((f) => onHideCallbacks.push(f())));
    }
    show._isDoing = false; // required to prevent infinite-calling from parent
    async function hide(hideCase, e) {
        show._isDoing && (await show._isDoing);
        if (!openedEl || hide._isDoing) {
            return;
        }
        const was = openedEl; // required when user clicks again during the hidding > we need to show in this case
        openedEl = null;
        onHideRef();
        try {
            hide._isDoing = true;
            const isOk = await onHide(hideCase, e || null);
            if (!isOk && !openedEl) {
                openedEl = was; // rollback if onHide was prevented and onShow wasn't called again during the hidding
                onShowCallbacks.forEach((f) => onHideCallbacks.push(f()));
            }
        }
        catch (error) {
            Promise.reject(error); // handle error from onHide
        }
        hide._isDoing = false;
    }
    hide._isDoing = false; // required to prevent infinite-calling from parent
    function appendEvent(...args) {
        const r = onEvent(...args);
        onRemoveCallbacks.push(r);
        return r;
    }
    // add event by popup.onShow and remove by onHide
    function onShowEvent(...args) {
        onShowCallbacks.push(() => onEvent(...args));
    }
    function includes(el) {
        return el instanceof Node && openedEl.contains(el);
    }
    // apply showCase
    let preventClickAfterFocus = false;
    let openedByHover = false;
    let debounceTimeout;
    // onClick
    if (opts.showCase & 4 /* ShowCases.onClick */) {
        let wasOutsideClick = false; // fix when labelOnClick > inputOnClick
        let lastActive = null;
        onShowEvent(document, "focusin", ({ target }) => {
            const isMe = openedEl === target || includes(target);
            /* istanbul ignore else */
            if (!isMe) {
                lastActive = target;
            }
        });
        let wasMouseMove = false; // fix when user makes t.mousedown, mousemove, body.mouseup
        appendEvent(t, "mousedown", (ev) => {
            wasMouseMove = false;
            if (ev.button) {
                // only for leftButton === 0
                return;
            }
            const r = onEvent(t, "mousemove", (e) => (wasMouseMove = e.movementX !== 0 || e.movementY !== 0), // it's required otherwise browser can provide move with 0 0 changes
            { once: true, passive: true });
            document.addEventListener("mouseup", () => r(), { once: true, passive: true });
        });
        onShowEvent(document, "click", (e) => {
            preventClickAfterFocus = false; // mostly it doesn't make sense but maybe it's possible
            if (wasMouseMove || e.detail === 2 || e.button) {
                wasMouseMove = false;
                return;
            }
            // filter click from target because we have target event for this
            const isTarget = t === e.target || (e.target instanceof Node && t.contains(e.target));
            if (!isTarget) {
                const isMeClick = openedEl === e.target || includes(e.target);
                if (isMeClick) {
                    focusFirst(lastActive || t);
                    hide(4 /* HideCases.onPopupClick */, e);
                }
                else {
                    hide(3 /* HideCases.onOutsideClick */, e);
                    wasOutsideClick = true;
                    setTimeout(() => (wasOutsideClick = false), 50);
                }
            }
        });
        appendEvent(t, "click", (e) => {
            if (!e.pageX) {
                // pageX is null or 0 if it was called programmatically
                preventClickAfterFocus = false; // test-case: focus without click > show....click programatically on target > it should hide
            }
            const skip = preventClickAfterFocus ||
                debounceTimeout ||
                wasOutsideClick ||
                openedByHover ||
                e.detail === 2 || // it's double-click
                e.button || // it's not left-click
                wasMouseMove;
            if (!skip) {
                if (!openedEl) {
                    lastActive = document.activeElement;
                    show(4 /* ShowCases.onClick */, e);
                }
                else {
                    const isMeClick = openedEl === e.target || includes(e.target);
                    hide(isMeClick ? 4 /* HideCases.onPopupClick */ : 5 /* HideCases.onTargetClick */, e);
                }
            }
            // fix when labelOnClick > inputOnClick > inputOnFocus
            debounceTimeout = setTimeout(() => (debounceTimeout = undefined), 1);
        });
    }
    // onHover
    if (opts.showCase & 1 /* ShowCases.onHover */) {
        let timeoutId;
        const ev = (ms, isMouseIn, e) => {
            timeoutId && clearTimeout(timeoutId);
            openedByHover = isMouseIn;
            if ((isMouseIn && !openedEl) || (!isMouseIn && openedEl))
                timeoutId = setTimeout(() => t.isConnected && // possible when target removed via innerHTML
                    (isMouseIn
                        ? show(1 /* ShowCases.onHover */, e) //
                        : hide(1 /* HideCases.onMouseLeave */, e)), ms);
            else
                timeoutId = undefined;
        };
        appendEvent(t, "mouseenter", (e) => ev(opts.hoverShowTimeout, true, e));
        // use only appendEvent; with onShowEvent it doesn't work properly (because filtered by timeout)
        appendEvent(t, "mouseleave", (e) => ev(opts.hoverHideTimeout, false, e));
    }
    // onFocus
    if (opts.showCase & 2 /* ShowCases.onFocus */) {
        const onFocused = async (e) => {
            if (!openedEl || debounceTimeout) {
                preventClickAfterFocus = !!(opts.showCase & 4 /* ShowCases.onClick */);
                await show(2 /* ShowCases.onFocus */, e);
                if (preventClickAfterFocus) {
                    const rst = () => {
                        preventClickAfterFocus = false;
                        r1();
                        r2();
                    };
                    const r1 = appendEvent(document, "touchstart", rst); // mousdown isn't not called when user touch-move-end
                    const r2 = appendEvent(document, "mousedown", rst);
                }
            }
        };
        onRemoveCallbacks.push(onFocusGot(t, onFocused, { debounceMs: opts.focusDebounceMs }));
        const blur = async (e) => {
            /* istanbul ignore else */
            if (openedEl) {
                const isToMe = openedEl === document.activeElement || openedEl === e.relatedTarget;
                const isToMeInside = !isToMe && includes(document.activeElement || e.relatedTarget);
                !isToMe && !isToMeInside && (await hide(2 /* HideCases.onFocusOut */, e));
                /* istanbul ignore else */
                if (!openedEl) {
                    openedByHover = false;
                }
            }
        };
        onShowCallbacks.push(() => onFocusLost(t, blur, { debounceMs: opts.focusDebounceMs }));
        const a = document.activeElement;
        if (!opts.skipAlreadyFocused && (a === t || (a instanceof HTMLElement && t.contains(a)))) {
            // isAlreadyFocused
            onFocused(new FocusEvent("focus"));
            preventClickAfterFocus = false;
        }
    }
    return { stopListen, hide, show };
}
popupListen.$defaults = {
    showCase: 4 /* ShowCases.onClick */,
    hoverShowTimeout: 200,
    hoverHideTimeout: 500,
};
